/******************************************************************************
 * HI 5323 Image Processing
 *
 * fourierview.cpp
 ******************************************************************************/

// stdlib includes
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
using namespace std;
// qt includes
#include <qapplication.h>
#include <qevent.h>
#include <qfont.h>
#include <qpainter.h>
// #include <qpixmap.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qlayout.h>
// #include <QMouseEvent>
// project includes
#include <fourierview.h>

// Maximum value.
inline float MAX(float a, float b)
{
    return a > b ? a : b;
};

/**
 * Destructor
 */
FourierView::~FourierView()
{
    delete m_pData;
    delete m_pFourier;
    delete m_pFourierTmp;
    delete m_pRed;
    delete m_pGreen;
    delete m_pBlue;
    delete m_pShow;
};

/**
 *
 */
void FourierView::_convert(int band, int show)
{
    uchar *row;
    int min, max;
    uint i, add, num;

    if (m_pData == NULL)
        return;

    // Scale factor to scale converted Complex value to 0-255.
    float   s, mag;
    QRgb   col;

    Complex *src;

    if (band >= RedBand && band <= BluBand)
    {
        if (m_bLockRGB)
        {
            min = RedBand;
            max = BluBand;
        } else
            min = max = band;

        max++;

    } else
        // Do not convert if no band specified.
        min = max = NoBand;

    for (band = min; band < max; band++)
    {
        src = m_pData;

        switch (band)
        {
        case RedBand : col = qRgba(1, 0, 0, 0);  break;
        case GrnBand : col = qRgba(0, 1, 0, 0);  break;
        case BluBand : col = qRgba(0, 0, 1, 0);  break;
        default      : col = qRgba(0, 0, 0, 0);  break;
        }

        row = (uchar *)&col;
        for (add = 0; row[add] == 0 && add < 4; add++);
        if (add == 4) return;

        num =m_iW *m_iH;  row = m_pFourierTmp->bits() + add;
        switch (show)
        {
        case ShowNone :
            for (i = num; i--; row += 4) *row = 0;
            break;

        case ShowReal :
            s = 255.0 / (m_fMaxReal -m_fMinReal);
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * (src->real() -m_fMinReal) + 0.5);
            break;

        case ShowImag :
            s = 255.0 / (m_fMaxImag -m_fMinImag);
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * (src->imag() -m_fMinImag) + 0.5);
            break;

        case ShowAbsReal :
            s = 255.0 / MAX(fabs(m_fMaxReal), fabs(m_fMinReal));
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * fabs(src->real()) + 0.5);
            break;

        case ShowAbsImag :
            s = 255.0 / MAX(fabs(m_fMaxImag), fabs(m_fMinImag));
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * fabs(src->imag()) + 0.5);
            break;

        case ShowLnReal_1 :
            s = 255.0 / log(m_fMaxReal -m_fMinReal + 1.0);
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * log(src->real() -m_fMinReal + 1.0) + 0.5);
            break;

        case ShowLnImag_1 : s = 255.0 / log(m_fMaxImag -m_fMinImag + 1.0);
        for (i = num; i--; src++, row += 4)
            *row = (uchar)(s * log(src->imag() -m_fMinImag + 1.0) + 0.5);
        break;

        case ShowLnAbsReal_1 :
            s = 255.0 / log(MAX(fabs(m_fMaxReal), fabs(m_fMinReal)) + 1.0);
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * log(fabs(src->real()) + 1.0) + 0.5);
            break;

        case ShowLnAbsImag_1 :
            s = 255.0 / log(MAX(fabs(m_fMaxImag), fabs(m_fMinImag)) + 1.0);
            for (i = num; i--; src++, row += 4)
                *row = (uchar)(s * log(fabs(src->imag()) + 1.0) + 0.5);
            break;

        case ShowMag :
            s = 255.0 / sqrt(m_fMaxMag);
            for (i = num; i--; src++, row += 4) {
                mag = src->real() * src->real() + src->imag() * src->imag();
                *row = (uchar)(s * sqrt(mag) + 0.5);
            }
            break;

        case ShowMag2 :
            s = 255.0 / m_fMaxMag;
            for (i = num; i--; src++, row += 4) {
                mag = src->real() * src->real() + src->imag() * src->imag();
                *row = (uchar)(s * mag + 0.5);
            }
            break;

        case ShowLnMag_1 :
            s = 255.0 / log(sqrt(m_fMaxMag) + 1.0);
            for (i = num; i--; src++, row += 4) {
                mag = src->real() * src->real() + src->imag() * src->imag();
                *row = (uchar)(s * log(sqrt(mag) + 1.0) + 0.5);
            }
            break;

        case ShowLnMag2_1 :
            s = 255.0 / log(m_fMaxMag + 1.0);
            for (i = num; i--; src++, row += 4) {
                mag = src->real() * src->real() + src->imag() * src->imag();
                *row = (uchar)(s * log(mag + 1.0) + 0.5);
            }
            break;

        case ShowPhase :
            s = 255.0 / M_2PI;
            for (i = num; i--; src++, row += 4) {
                mag = atan2(src->imag(), src->real());
                if (mag < 0.0) mag += M_2PI;
                *row = (uchar)(s * mag + 0.5);
            }
            break;

        default :
            cerr << "FourierView> ERROR: unknown conversion requested." << endl;
            exit(0);
        }
    }

    if (m_bRecenter)
    {
        uint x = m_iW >> 1, y = m_iH >> 1, w = m_iW - x, h = m_iH - y;

	QPixmap oPixmap;
        oPixmap.convertFromImage( *m_pFourierTmp );
        QPainter painter(m_pFourier);
        painter.drawPixmap(0, 0, oPixmap, w, h, x, y);
        painter.drawPixmap(x, 0, oPixmap, 0, h, w, y);
        painter.drawPixmap(0, y, oPixmap, w, 0, x, h);
        painter.drawPixmap(x, y, oPixmap, 0, 0, w, h);
    } else {
        m_pFourier->convertFromImage( *m_pFourierTmp );
    }

    repaint();
}

void FourierView::initGui()
{
    // main groupbox
    m_pShow = new GroupBox("Fourier Spectrum Display Options");
    m_pShow->setColumnLayout( 2, Qt::Vertical );

    // options
    GroupBox *pFlags  = new GroupBox("Display options", m_pShow);
    pFlags->setColumnLayout( 2, Qt::Horizontal );
    QCheckBox* pCenter = new QCheckBox ("Recenter spectrum", pFlags);
    QCheckBox* pLock   = new QCheckBox ("Lock RGB bands",    pFlags);

    // bands
    GroupBox *pBands = new GroupBox("Complex display conversion: z = a + bi", m_pShow);
    pBands->setColumnLayout( 3, Qt::Horizontal );

    // red band
    m_pRed = new ButtonGroup( "Red", pBands );
    m_pRed->setExclusive( true );
    m_pRed->setColumnLayout(15, Qt::Vertical );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );
    new QRadioButton( "---", (QWidget*)m_pRed );

    // green band
    m_pGreen = new ButtonGroup( "Green", pBands );
    m_pGreen->setExclusive( true );
    m_pGreen->setColumnLayout(15, Qt::Vertical );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );
    new QRadioButton( "---", (QWidget*)m_pGreen );

    // blue band
    m_pBlue = new ButtonGroup( "Blue", pBands );
    m_pBlue->setExclusive( true );
    m_pBlue->setColumnLayout(15, Qt::Vertical );
    new QRadioButton( "Nothing", (QWidget*)m_pBlue );
    new QRadioButton( "Real [a]", (QWidget*)m_pBlue );
    new QRadioButton( "Imag [b]", (QWidget*)m_pBlue );
    new QRadioButton( "|Real| [fabs(a)]", (QWidget*)m_pBlue );
    new QRadioButton( "|Imag| [fabs(b)]", (QWidget*)m_pBlue );
    new QRadioButton( "ln(Real + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "ln(Imag + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "ln(|Real| + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "ln(|Imag| + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "Mag [sqrt(a*a + b*b)]", (QWidget*)m_pBlue );
    new QRadioButton( "Mag_Sqrd [a*a + b*b]", (QWidget*)m_pBlue );
    new QRadioButton( "ln(Mag + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "ln(Mag_Sqrd + 1)", (QWidget*)m_pBlue );
    new QRadioButton( "Phase [arctan(b/a)]", (QWidget*)m_pBlue );

    m_bRecenter = TRUE;  m_bLockRGB = TRUE;

    connect(pCenter, SIGNAL(toggled(bool)),       SLOT(setCenter (bool)));
    connect(pLock,   SIGNAL(toggled(bool)),       SLOT(setLockRGB(bool)));

    connect(m_pRed,  SIGNAL(clicked(int )),       SLOT(redChanged(int )));
    connect(m_pGreen,SIGNAL(clicked(int )),       SLOT(grnChanged(int )));
    connect(m_pBlue, SIGNAL(clicked(int )),       SLOT(bluChanged(int )));
    connect(this,    SIGNAL(pressed(    )), qApp, SLOT(      quit(    )));

    setMouseTracking(TRUE);

    setBackgroundColor(QColor(0,0,0));

    m_pShow->show();

    m_pRed->setButton(ShowLnMag_1);
    m_pGreen->setButton(ShowLnMag_1);
    m_pBlue->setButton(ShowLnMag_1);

#ifdef QT3_SUPPORT
    pCenter->setCheckState(Qt::Checked);
    pLock->setCheckState(Qt::Checked);
#endif
#ifndef QT3_SUPPORT
    pCenter->setChecked(true);
    pLock->setChecked(true);
#endif

}

void FourierView::initData()
{
    uint i;
    float mag;
    Complex z;

    m_fMinReal = m_fMaxReal = m_pData->real();
    m_fMinImag = m_fMaxImag = m_pData->imag();
    m_fMaxMag  = (m_fMaxReal * m_fMaxReal) + (m_fMaxImag *m_fMaxImag);

    for (i=0; i < m_iW * m_iH; i++)
    {
	z = m_pData[i];

	if (z.real() >m_fMaxReal)
	    m_fMaxReal = z.real();
	else if (z.real() <m_fMinReal)
	    m_fMinReal = z.real();

	if (z.imag() >m_fMaxImag)
	    m_fMaxImag = z.imag();
	else if (z.imag() <m_fMinImag)
	    m_fMinImag = z.imag();

	mag = z.real() * z.real() + z.imag() * z.imag();
	if (mag >m_fMaxMag)
	    m_fMaxMag = mag;
    }

    _convert(RedBand, ShowLnMag_1);
    _convert(GrnBand, ShowLnMag_1);
    _convert(BluBand, ShowLnMag_1);

    setMinimumSize(m_iW*2,m_iH);
    setMaximumSize(m_iW*2,m_iH);
};

FourierView::FourierView(Complex *pData, QImage* pImage, QWidget *pParent, const char *name) : QWidget(pParent, name)
{
    m_pData = pData;
    m_pImage = pImage;
    m_iW = m_pImage->width();
    m_iH = m_pImage->height();

    m_pFourier = new QPixmap( (int)m_iW, (int)m_iH );
    m_pFourierTmp = new QImage( (int)m_iW, (int)m_iH, 32 );
    m_pFourierTmp->setNumColors( 256 );
    for (uint j = 0; j < 256; j++)
	m_pFourierTmp->setColor(j, qRgb(j, j, j));

    cout << "FourierView> Image size: " << m_iW << "x" << m_iH << endl;

    initData();
    initGui();
};

void FourierView::paintEvent(QPaintEvent *)
{
    QPainter *paint;

    paint = new QPainter();
    paint->begin(this);

    // paint original image
    if (m_pImage != NULL)
	paint->drawImage(0,0,*m_pImage);
    else
	cout << "No original image!" << endl;

    // paint fourier image
    if (m_pFourier != NULL)
        paint->drawPixmap((int)m_iW,0,*m_pFourier);
    else
        cout << "No pixmap!" << endl;

    paint->end();
}


void FourierView::setCenter(bool center)
{
    m_bRecenter = center;
    _convert(NoBand, ShowNone);
};

void FourierView::setLockRGB(bool lock)
{
    if (lock)
    {
        int id = m_pGreen->selectedId();

        if (m_pRed->selectedId() != id)
        {
	    m_pRed->setButton(id);
            _convert(RedBand, id);
        }
        if (m_pBlue->selectedId() != id)
        {
            m_pBlue->setButton(id);
            _convert(BluBand, id);
        }
    }
    m_bLockRGB = lock;
};

void FourierView::redChanged(int show)
{
    if (m_bLockRGB)
    {
	m_pGreen->setButton(show);
	m_pBlue->setButton(show);
    }
    _convert(RedBand, show);
};

void FourierView::grnChanged(int show)
{
    if (m_bLockRGB)
    {
	m_pRed->setButton(show);
	m_pBlue->setButton(show);
    }
    _convert(GrnBand, show);
};

void FourierView::bluChanged(int show)
{
    if (m_bLockRGB)
    {
	m_pRed->setButton(show);
	m_pGreen->setButton(show);
    }
    _convert(BluBand, show);
};

void FourierView::mousePressEvent(QMouseEvent *)
{
  emit pressed();
};

void FourierView::mouseMoveEvent(QMouseEvent *me)
{
  int x = me->pos().x(), y = me->pos().y();
  Complex z;

  if (x < (int)m_iW || x >= (int)(m_iW*2) || y < 0 || y >= (int)m_iH)
      return;

  x -= m_iW;

  if (m_bRecenter)
  {
      x -= m_iW >> 1;
      y -= m_iH >> 1;

      if (x < 0) x
          += m_iW;
      if (y < 0) y
          += m_iW;
  }
  z = m_pData[x + (y * m_iW)];

  printf( "[%3d,%3d] -> %10.6f + %10.6fi -> Mag: %10.6f  Phase: %7.4f pi\n", x, y, z.real(), z.imag(), z.abs(), z.arg() / M_2PI);
}
